from youtube_autonomous.database.database_handler import DatabaseHandler
from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.rules.effect_element_rules import ElementRules
from youtube_autonomous.elements.rules.rules_checker import RulesChecker
from youtube_autonomous.segments.builder.config import DEFAULT_SEGMENT_PARTS_FOLDER
from yta_multimedia.audio.voice.transcription.stt.whisper import get_transcription_with_timestamps
from yta_general_utils.file import copy_file
from yta_general_utils.file.filename import get_file_extension
from yta_general_utils.temp import get_temp_filename
from yta_general_utils.logger import print_completed
from moviepy.editor import AudioFileClip, VideoFileClip


class Enhancement:
    project_id: int
    segment_index: int
    segment = None
    index: int
    status = ''
    type = ''

    _audio_filename: str
    audio_clip = None
    _video_filename: str
    video_clip = None
    _full_filename: str
    full_clip = None

    audio_narration_filename: str
    voice: str
    narration_text: str

    text: str

    keywords: str

    start: float = None
    duration: float = None

    filename: str
    url: str

    mode: str

    music: str

    enhancements: list

    narration_text_sanitized_without_shortcodes: str
    narration_text_with_simplified_shortcodes: str
    narration_text_sanitized: str
    transcription = None
    shortcodes = None
    created_at = None

    @property
    def audio_filename(self):
        return self._audio_filename

    @audio_filename.setter
    def audio_filename(self, audio_filename):
        self._audio_filename = audio_filename

        # TODO: Check is valid or reset (in database) and raise
        # Exception (?)
        if self.audio_filename:
            self.audio_clip = AudioFileClip(self.audio_filename)

    @property
    def video_filename(self):
        return self._video_filename
    
    @video_filename.setter
    def video_filename(self, video_filename):
        self._video_filename = video_filename

        # TODO: Check is valid or reset (in database) and raise
        # Exception (?)
        if self.video_filename:
            self.video_clip = VideoFileClip(self.video_filename)

    @property
    def full_filename(self):
        return self._full_filename
    
    @full_filename.setter
    def full_filename(self, full_filename):
        self._full_filename = full_filename

        # TODO: Check is valid or reset (in database) and raise
        # Exception (?)
        if self.full_filename:
            self.video_clip = VideoFileClip(self.full_filename)

    def __init__(self, project_id, segment_index: int, segment, index: int, data: dict):
        self.project_id = str(project_id)
        self.segment_index = segment_index
        self.segment = segment
        self.index = index

        for key in data:
            setattr(self, key, data[key])

        self.rules = ElementRules.get_subclass_by_type(data['type'])()
        self.builder = ElementBuilder.get_subclass_by_type(data['type'])()
        self.rules_checker = RulesChecker(self.rules)

        self.database_handler = DatabaseHandler()

    # TODO: This below has been moved because inheritance is working strangely

    def create_segment_file(self, filename: str):
        """
        Creates a filename within the definitive segments folder
        to keep the generated file locally to recover it in the
        next project execution if something goes wrong. The 
        definitive filename will be built using the provided
        'filename' and adding some more information in the name.

        This method will generate a temporary filename that uses
        the current segment index in its name and is placed in 
        segment parts folder.

        This method returns the final filename created.
        """
        if not filename:
            raise Exception('No "filename" provided.')

        temp_filename = get_temp_filename(filename)

        return f'{DEFAULT_SEGMENT_PARTS_FOLDER}/segment_{self.index}_{temp_filename}'
    
    def create_narration(self):
        """
        Creates the audio narration (if needed) by generating an AI audio
        narration with provided 'voice' and 'audio_narration_text'
        parameters or by using the 'audio_narration_filename'.

        This method will set the 'audio_filename' to be able to build the
        audio clip in a near future.
        """
        if self.audio_narration_filename:
            segment_part_filename = self.create_segment_file(f'narration.{get_file_extension(self.audio_narration_filename)}')
            copy_file(self.audio_narration_filename, segment_part_filename)
            print_completed('Original voice narration file copied to segment parts folder')
            self.audio_narration_filename = segment_part_filename
            self.audio_filename = segment_part_filename
        else:
            segment_part_filename = self.create_segment_file('narration.wav')
            # TODO: Voice parameter need to change
            self.audio_filename = self.builder.build_narration(self.narration_text_sanitized_without_shortcodes, output_filename = segment_part_filename)
            print_completed('Voice narration created successfully')

    def create_transcription(self):
        """
        Creates the transcription of the generated audio narration
        that would be stored in 'self.audio_filename'.
        
        This method returns a words array containing, for each word,
        a 'text', 'start' and 'end' field to be able to use the 
        transcription timestamps.
        """
        transcription = get_transcription_with_timestamps(self.audio_filename, initial_prompt = self.narration_text_sanitized_without_shortcodes)

        # We join all words together as we don't need segments, just words with time
        words = []
        for transcription_segment in transcription['segments']:
            words += transcription_segment['words']
        self.transcription = words

    # TODO: This above has been moved because inheritance is working strangely

    # TODO: This below has to be in the common Object from wich Segment
    # and Enhancement will inherit
    def build_step_1_create_narration(self):
        # 1. Generate narration if needed
        self.create_narration()

    def build_step_2_build_base_content(self):
        # 2. Build base video
        if not self.video_clip:
            print(self.segment)
            self.video_clip = self.builder.build_from_enhancement(self, self.segment)
            filename = self.create_segment_file('video.mp4')
            self.video_clip.write_videofile(filename)
            self.video_filename = filename

        # TODO: In the future we could add enhancements to this
        if self.audio_clip:
            self.video_clip = self.video_clip.set_audio(self.audio_clip)
        self.full_clip = self.video_clip
        filename = self.create_segment_file('video.mp4')
        self.full_clip.write_videofile(filename)
        self.full_filename = filename
    
    def build(self):
        """
        Builds this enhancement to be applied in the provided 'segment'.
        """
        self.build_step_1_create_narration()
        self.build_step_2_build_base_content()

        return self.full_clip


from youtube_autonomous.segments.enums import SegmentStatus, SegmentBuildingField, ProjectStatus, ProjectField, SegmentField
from youtube_autonomous.elements.validator.element_parameter_validator import ParameterValidator
from youtube_autonomous.shortcodes.shortcode_parser import ShortcodeParser
from bson.objectid import ObjectId
from pymongo import MongoClient
from datetime import datetime
from typing import Union


class DatabaseHandler:
    """
    This class is used to interact with the database in which the
    projects are stored and its information and status is handled.

    The projects are stored in the database containing the mongo
    '_id'', a 'status', the 'script' (that is the whole json used
    to generate it and to compare other json files to check if
    existing), and the 'segments' field that is the one the app
    uses to handle the building data and this process.

    The 'script' must be preserved as it is, to be able to compare
    jsons and avoid duplicated projects, and the 'segments' field
    must be used by the app to keep the building process progress.
    """
    DATABASE_NAME = 'youtube_autonomous'
    CONNECTION_STRING = 'mongodb://localhost:27017/?readPreference=primary&appname=MongoDB%20Compass&directConnection=true&ssl=false/' + DATABASE_NAME  # previous: youtubeautonomo
    client = None

    def __init__(self):
        # Initialize database connection
        # TODO: I think this should be a Singleton
        self.get_database()

    def get_database(self):
        """
        Returns the client connected to the database.
        """
        if not self.client:
            self.client = MongoClient(self.CONNECTION_STRING)

        return self.client[self.DATABASE_NAME]
    
    def get_database_projects_table(self):
        """
        Returns the database 'projects' table
        """
        return self.get_database()['projects']
    
    def get_unfinished_project(self):
        """
        Returns the first unfinished project existing in database,
        or None if there are no unfinished projects.

        TODO: Is it actually the first or the last one (?)
        """
        db_projects_table = self.get_database_projects_table()

        db_unfinished_projects = db_projects_table.find({
            "status": { "$ne": ProjectStatus.FINISHED.value }
        })

        try:
            return db_unfinished_projects.next()
        except:
            return None
    
    def get_unfinished_projects(self):
        """
        Returns the unfinished projects that exist in the
        database or an empty array if there are no unfinished
        projects.
        """
        db_projects_table = self.get_database_projects_table()

        db_unfinished_projects = db_projects_table.find({
            "status": { "$ne": ProjectStatus.FINISHED.value }
        })

        # Build a projects array
        # TODO: Improve this, please
        projects = []
        do_finish = False
        while not do_finish:
            try:
                projects.append(db_unfinished_projects.next())
            except:
                do_finish = True
                pass

        return projects
    
    # TODO: Apply 'json' type
    def get_database_project_from_json(self, json):
        """
        Returns the project with the provided 'json' that is stored
        in the database if it actually exists. The 'json' is the 
        whole script that builds the video segments.
        """
        db_projects = self.get_database_projects_table()

        db_project = db_projects.find({
            'script': json
        })

        # TODO: Improve this if possible
        try:
            return db_project.next()
        except:
            return None
        
    # TODO: Apply the 'id' type
    def get_database_project_from_id(self, db_project_id: Union[ObjectId, str]):
        """
        Returns the project with the given 'id' if available, or None
        if not.
        """
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)

        db_projects = self.get_database_projects_table()

        db_project = db_projects.find({
            '_id': db_project_id
        })

        # TODO: Improve this if possible
        try:
            return db_project.next()
        except:
            return None
        
    def insert_project(self, data: dict):
        """
        Creates a new project with the provided 'data' structure
        that is made of 'status', 'script' and 'segments' fields.

        The 'data' provided here is always valid so I don't check 
        it.
        """
        # TODO: Make some checkings (?)
        db_projects_table = self.get_database_projects_table()
        db_project_id = db_projects_table.insert_one(data).inserted_id
        
        return self.get_database_project_from_id(db_project_id)
    
    def update_project_segment_field(self, db_project_id: Union[ObjectId, str], segment_index: int, field: Union[SegmentField, SegmentBuildingField, str], value):
        """
        Updates a segment field from the provided project (identified by 
        'project_id').
        """
        if not segment_index and segment_index != 0:
            raise Exception('No "segment_index" provided.')

        if segment_index < 0:
            raise Exception('In valid "segment_index" provided')

        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)

        if not field:
            raise Exception('No "field" provided.')
        
        if not isinstance(field, (SegmentField, SegmentBuildingField, str)):
            raise Exception('The "field" parameter is not a SegmentField nor a SegmentBuildingField nor a string.')
        
        # I let this 'field' to be not strict and accept any string
        if isinstance(field, (SegmentField, SegmentBuildingField)):
            field = field.value
        
        # TODO: Improve checkings
        query = { '_id': db_project_id }
        new_values = { "$set": { 'segments.' + str(segment_index) + '.' + field: value } }
        # TODO: Do you see this below more clear than the one above (?)
        #new_values = { "$set": { f'segments.{str(segment_index)}.{field: value}' } }

        db_projects_table = self.get_database_projects_table()
        db_projects_table.update_one(query, new_values)

    def update_project_segment_enhancement_field(self, db_project_id: Union[ObjectId, str], segment_index: int, enhancement_index: int, field: Union[SegmentField, str], value):
        ParameterValidator.validate_mandatory_parameter('db_project_id', db_project_id)
        ParameterValidator.validate_numeric_positive_mandatory_parameter('segment_index', segment_index)
        ParameterValidator.validate_numeric_positive_mandatory_parameter('enhancement_index', enhancement_index)
        # TODO: Only these field type enums (?)
        ParameterValidator.validate_is_class('field', field, [SegmentField, SegmentBuildingField, str])
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)

        if isinstance(field, (SegmentField, SegmentBuildingField)):
            field = field.value

        # TODO: Improve checkings
        query = { '_id': db_project_id }
        field_str = f'segments.{str(segment_index)}.enhancements.{str(enhancement_index)}.{field}'
        new_values = { "$set": {field_str: value} }

        db_projects_table = self.get_database_projects_table()
        db_projects_table.update_one(query, new_values)

    def update_project_segment_status(self, db_project_id: Union[ObjectId, str], segment_index: int, status: Union[SegmentStatus, str]):
        """
        Updates, in the project with the provided 'db_project_id'
        id, the segment in 'segment_index' index, setting the also
        provided 'status'.
        """
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)
        
        if not segment_index and segment_index != 0:
            raise Exception('No "segment_index" provided.')
        
        if not status:
            raise Exception('No "status" provided.')

        if not isinstance(status, (SegmentStatus, str)):
            raise Exception('The "status" parameter provided is not a SegmentStatus nor a string.')
        
        if isinstance(status, str):
            if not SegmentStatus.is_valid(status):
                raise Exception('The "status" provided is not a valid SegmentStatus value.')
            
            status = SegmentStatus(status)
        
        self.update_project_segment_field(db_project_id, segment_index, ProjectField.STATUS.value, status.value)

    def set_project_segment_as_in_progress(self, db_project_id: Union[ObjectId, str], segment_index: int):
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)
        
        if not segment_index and segment_index != 0:
            raise Exception('No "segment_index" provided.')
        
        self.update_project_segment_status(db_project_id, segment_index, SegmentStatus.IN_PROGRESS.value)

    def set_project_segment_as_finished(self, db_project_id: Union[ObjectId, str], segment_index: int):
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)
        
        if not segment_index and segment_index != 0:
            raise Exception('No "segment_index" provided.')
        
        self.update_project_segment_status(db_project_id, segment_index, SegmentStatus.FINISHED)
        
        # TODO: Maybe SegmentBuildingField.FINISHED_AT (?)
        self.update_project_segment_field(db_project_id, segment_index, 'finished_at', datetime.now())

    def set_project_segment_transcription(self, db_project_id: Union[ObjectId, str], segment_index: int, transcription):
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)
        
        if not segment_index and segment_index != 0:
            raise Exception('No "segment_index" provided.')
        
        self.update_project_segment_field(db_project_id, segment_index, SegmentBuildingField.TRANSCRIPTION.value, transcription)

    def set_project_segment_audio_filename(self, db_project_id: Union[ObjectId, str], segment_index: int, audio_filename):
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)
        
        if not segment_index and segment_index != 0:
            raise Exception('No "segment_index" provided.')
        
        self.update_project_segment_field(db_project_id, segment_index, SegmentBuildingField.AUDIO_FILENAME.value, audio_filename)

    def set_project_segment_video_filename(self, db_project_id: Union[ObjectId, str], segment_index: int, video_filename):
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)
        
        if not segment_index and segment_index != 0:
            raise Exception('No "segment_index" provided.')
        
        self.update_project_segment_field(db_project_id, segment_index, SegmentBuildingField.VIDEO_FILENAME.value, video_filename)

    def set_project_segment_full_filename(self, db_project_id: Union[ObjectId, str], segment_index: int, full_filename):
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)
        
        if not segment_index and segment_index != 0:
            raise Exception('No "segment_index" provided.')
        
        self.update_project_segment_field(db_project_id, segment_index, SegmentBuildingField.FULL_FILENAME.value, full_filename)

    def update_project_field(self, db_project_id: Union[ObjectId, str], field: Union[ProjectField, str], value):
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)

        if not field:
            raise Exception('No "field" provided.')
        
        if not isinstance(field, (ProjectField, str)):
            raise Exception('The "field" parameter provided is not a ProjectField nor a string.')
        
        if isinstance(field, ProjectField):
            field = field.value
        
        db_project = self.get_database_project_from_id(db_project_id)

        if not db_project:
            return None
        
        db_projects_table = self.get_database_projects_table()
        query = { '_id': db_project_id }
        new_values = { "$set": { field: value } }
        db_projects_table.update_one(query, new_values)

    def set_project_music_filename(self, db_project_id: Union[ObjectId, str], music_filename: str):
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)

        # TODO: Maybe SegmentBuildingField.MUSIC_FILENAME (?)
        self.update_project_field(db_project_id, 'music_filename', music_filename)

    def set_project_as_in_progress(self, db_project_id: Union[ObjectId, str]):
        """
        Sets the project (identified by 'db_project_id') status as 
        'in_progress'.
        """
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)

        self.update_project_field(db_project_id, ProjectField.STATUS, ProjectStatus.IN_PROGRESS.value)

    def set_project_as_finished(self, db_project_id: Union[ObjectId, str]):
        """
        Sets the project (identified by 'db_project_id') status as
        'finished'.
        """
        if not db_project_id:
            raise Exception('No "db_project_id" provided.')
        
        if isinstance(db_project_id, str):
            db_project_id = ObjectId(db_project_id)

        self.update_project_field(db_project_id, ProjectField.STATUS, ProjectStatus.FINISHED.value)



from youtube_autonomous.segments.enums import SegmentField, ElementBuildingField
from youtube_autonomous.elements.enhancement import Enhancement
from youtube_autonomous.elements.validator.element_validator import StringDuration
from youtube_autonomous.elements.rules.effect_element_rules import ElementRules
from youtube_autonomous.elements.rules.rules_checker import RulesChecker
from youtube_autonomous.shortcodes.objects.shortcode import Shortcode
from youtube_autonomous.segments.enhancement.edition_manual.edition_manual import EditionManual
from youtube_autonomous.elements.element import Element
from yta_general_utils.logger import print_completed
from bson.objectid import ObjectId
from typing import Union


class Segment(Element):
    @property
    def status(self):
        return super().status
    
    @status.setter
    def status(self, status: str):
        super().status = status

        if self._do_update_database:
            self.database_handler.update_project_segment_status(self.project_id, self.index, status)

    @property
    def audio_filename(self):
        return super().audio_filename
    
    @audio_filename.setter
    def audio_filename(self, audio_filename: Union[str, None]):
        audio_filename = audio_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, ElementBuildingField.AUDIO_FILENAME.value, audio_filename)

    @property
    def video_filename(self):
        return super().video_filename
    
    @video_filename.setter
    def video_filename(self, video_filename: Union[str, None]):
        video_filename = video_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, ElementBuildingField.VIDEO_FILENAME.value, video_filename)

    @property
    def full_filename(self):
        return super().full_filename
    
    @full_filename.setter
    def full_filename(self, full_filename: Union[str, None]):
        super().full_filename = full_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, ElementBuildingField.FULL_FILENAME.value, full_filename)

    @property
    def audio_narration_filename(self):
        return super().audio_narration_filename
    
    @audio_narration_filename.setter
    def audio_narration_filename(self, audio_narration_filename: Union[str, None]):
        super().audio_narration_filename = audio_narration_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, SegmentField.AUDIO_NARRATION_FILENAME.value, audio_narration_filename)

    @property
    def transcription(self):
        return super().transcription
    
    @transcription.setter
    def transcription(self, transcription: Union[dict, None]):
        transcription = transcription

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, ElementBuildingField.TRANSCRIPTION.value, transcription)

    @property
    def shortcodes(self):
        return super().shortcodes
    
    @shortcodes.setter
    def shortcodes(self, shortcodes: Union[list[Shortcode], None]):
        shortcodes = shortcodes

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, ElementBuildingField.SHORTCODES.value, shortcodes)

    @property
    def calculated_duration(self):
        return super().calculated_duration
    
    @calculated_duration.setter
    def calculated_duration(self, calculated_duration: Union[float, int, None]):
        calculated_duration = calculated_duration

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, ElementBuildingField.CALCULATED_DURATION.value, calculated_duration)

    def __init__(self, project_id: Union[str, ObjectId], index: int, data: dict):
        super().__init__(index, data)

        self._do_update_database = False

        self.project_id = str(project_id)
        
        # TODO: This should go in 'self.data' (?)
        self.enhancements = []
        for enhancement_index, enhancement in enumerate(data.get('enhancements', [])):
            self.enhancements.append(Enhancement(self.project_id, self.index, enhancement_index, enhancement))

        self._do_update_database = True

    def validate(self):
        super().validate()

        # If 'enhancements' are invalid
        # TODO: self.enhancements doesn't exist yet
        for enhancement in self.enhancements:
            RulesChecker.check_need_rules(enhancement, ElementRules.get_subclass_by_type(enhancement['type']))

    # TODO: Remove this method below
    def prepare(self):
        # Check duration
        if self.data.duration and not self.calculated_duration:
            self.calculated_duration = self.data.duration

        # Generate narration audio
        if self.rules_checker.should_build_narration_rule(self.data) and not self.audio_clip:
            self.handle_narration()

        # Generate video
        if not self.video_clip:
            self.video_clip = self.builder.build_segment(self)
            # We write this video part to be able to recover it in the future
            segment_part_filename = self.create_segment_file('video.mp4')
            self.video_clip.write_videofile(segment_part_filename)
            self.video_filename = segment_part_filename
            print_completed('Created basic video part and stored locally and in database.')

    # TODO: This below has to be in the common Object from wich Segment
    # and Enhancement will inherit
    def build_step_1_create_narration(self):
        # 1. Generate narration if needed
        self.create_narration()

    def build_step_2_create_transcription(self):
        # 2. Generate narration transcription
        self.create_transcription()

    def build_step_2_extract_user_shortcodes(self):
        # 2. Extract manually written shortcodes in 'narration_text'
        self.shortcode_parser.parse(self.narration_text)
        self.shortcodes = self.shortcode_parser.shortcodes
        # TODO: Turn shortcodes into enhancements
        enhancements = [shortcode.to_enhancement_element(self.transcription) for shortcode in self.shortcodes]
        print(enhancements)

    def build_step_3_apply_edition_manual_shortcodes(self):
        # 3. Apply shortcodes from Edition Manual and 'narration_text'
        self.apply_edition_manual()

    def build_step_4_build_base_content(self):
        # 4. Build base video
        self.video_clip = self.builder.build_from_segment(self)
        filename = self.create_segment_file('video.mp4')
        self.video_clip.write_videofile(filename)
        self.video_filename = filename

    def build_step_5_update_enhancements_duration(self):
        # 5. Update enhancements duration according to base video
        # TODO: What if 'start' is after the end of the base video (?)
        for enhancement in self.enhancements:
            if enhancement.duration == StringDuration.SEGMENT_DURATION.name:
                enhancement.duration = self.video_clip.duration

            if enhancement.start >= self.video_clip.duration:
                raise Exception('The enhancement start moment is after the video_clip.duration')
            
            end = enhancement.start + enhancement.duration
            if end > self.video_clip.duration:
                enhancement.duration = self.video_clip.duration - enhancement.start
            
    def build_step_6_build_enhancements(self):
        for enhancement in self.enhancements:
            enhancement_clip = enhancement.build()
            # TODO: Combine with our segment according to type

    def build(self):
        if self.rules_checker.should_build_narration_rule(self):
            if not self.audio_filename:
                self.build_step_1_create_narration()

            # We are forcing duration here
            self.duration = self.audio_clip.duration

            if not self.transcription:
                self.build_step_2_create_transcription()

            # TODO: Handle this (maybe status (?))
            self.build_step_2_extract_user_shortcodes()

            self.build_step_3_apply_edition_manual_shortcodes()

        # TODO: What about duration here that is not set (?)
        if self.audio_clip:
            self.calculated_duration = self.audio_clip.duration

        print(self.video_clip)
        
        if not self.video_clip:
            print('Building base content step 4')
            self.build_step_4_build_base_content()

        # Make the base video to apply enhancements on it
        if self.audio_clip:
            self.video_clip = self.video_clip.set_audio(self.audio_clip)

        self.build_step_5_update_enhancements_duration()
        self.build_step_6_build_enhancements()

    def apply_edition_manual(self):
        # TODO: I need to dynamically get the edition manual from somewhere
        # By now I'm forcing it
        test_edition_manual = 'C:/Users/dania/Desktop/PROYECTOS/youtube-autonomous/youtube_autonomous/segments/enhancement/edition_manual/example.json'
        edition_manual = EditionManual.init_from_file(test_edition_manual)
        dict_enhancements_found = edition_manual.apply(self.transcription)
        # Turn dict enhancements found into Enhancement objects
        for index, dict_enhancement_found in enumerate(dict_enhancements_found):
            # TODO: What do we do with 'index' here (?)
            print(type(dict_enhancement_found))
            index = 100 + index
            self.enhancements.append(Enhancement(self.project_id, self.index, index, dict_enhancement_found))
        # TODO: Some enhancements could be incompatible due to collisions
        # or things like that

        # TODO: What about this enhancements (?)


  


from youtube_autonomous.elements.rules.element_rules import ElementRules
from youtube_autonomous.segments.enums import EnhancementElementMode


class StockElementRules(ElementRules):
    def __init__(self):
        self.can_have_narration = True
        self.need_narration = False
        self.can_have_specific_duration = True
        self.need_specific_duration = False
        self.can_have_text = False
        self.need_text = False
        self.can_have_filename = False
        self.can_have_url = False
        self.need_filename_or_url = False
        self.can_have_keywords = True
        self.need_keywords = True
        self.can_have_more_parameters = False

        self.can_be_segment = True
        self.can_be_enhancement_element = True
        self.valid_enhancement_modes = [EnhancementElementMode.INLINE, EnhancementElementMode.OVERLAY]
        self.default_enhancement_mode = EnhancementElementMode.OVERLAY


from youtube_autonomous.segments.enums import EnhancementElementField, EnhancementElementType, EnhancementElementDuration, EnhancementElementStart, EnhancementElementMode
from yta_general_utils.checker.type import variable_is_positive_number
from typing import Union


class SegmentEnhancementValidator:
    """
    This class is to validate the Segment enhancement elements that
    are terms that the user register to enhance (to improve) the
    project video experience.

    These terms need to have a valid structure and that's what we
    check here.
    """
    def is_valid(enhancement_terms):
        """
        We will receive the content of a enhancement term and will raise an
        Exception if some structure element or value is not valid according
        to our rules.

        TODO: Write a little bit more about what we are receiving here.
        """
        # TODO: Maybe refactor the terms to simplify them
        for enhancement_term_key in enhancement_terms:
            enhancement_term = enhancement_terms[enhancement_term_key]
            for type in enhancement_term:
                content = enhancement_term[type]
                if not EnhancementElementType.is_valid(type):
                    raise Exception('Enhancement element type "' + type + '" not accepted.')

                # Lets check values
                keywords = content.get(EnhancementElementField.KEYWORDS.value)
                filename = content.get(EnhancementElementField.FILENAME.value)
                url = content.get(EnhancementElementField.URL.value)
                start = content.get(EnhancementElementField.START.value)
                mode = content.get(EnhancementElementField.MODE.value)
                duration = content.get(EnhancementElementField.DURATION.value)

                if not keywords and not filename and not url:
                    raise Exception('No "keywords" nor "filename" nor "url" provided.')
                
                # TODO: Search for the element with 'keywords' to check that it
                # doesn't exist (or it does)
                # TODO: Check that provided 'filename' is a valid file and the
                # type is appropriate for the type
                # TODO: Check that the provided 'url' is valid, available and
                # suitable for the type we want
                
                if type == EnhancementElementType.MEME.value and not keywords:
                    raise Exception(f'No "keywords" provided when necessary for the "{EnhancementElementType.MEME.value}" type.')

                start_values = EnhancementElementStart.get_all_values()
                if start and start not in start_values and not variable_is_positive_number(start):
                    
                    raise Exception(f'No valid "start" value provided. Must be one of these "{EnhancementElementStart.get_all_values_as_str()}" or a valid positive number (including 0).')

                duration_values = EnhancementElementDuration.get_all_values()
                if duration and duration not in duration_values and not variable_is_positive_number(duration):
                    raise Exception(f'No valid "duration" value provided. Must be one of these "{EnhancementElementDuration.get_all_values_as_str()}" or a valid positive number (including 0).')

                mode_values = EnhancementElementMode.get_all_values()
                if mode and mode not in mode_values:
                    raise Exception(f'No valid "mode" value provided. Must be one of these "{EnhancementElementMode.get_all_values_as_str()}".')
                
    @classmethod
    def validate_type(cls, type: Union[EnhancementElementType, str]):
        """
        This method validates if the provided 'type' parameter is an
        EnhancementElementType enum or if it is a valid string value 
        of a EnhancementElementType enum.

        This method will raise an Exception if any of the rules are 
        not fit.
        """
        if not type:
            raise Exception('No "type" provided.')
        
        if not isinstance(type, (EnhancementElementType, str)):
            raise Exception(f'The "type" parameter provided {type} is not an EnhancementElementType nor a string.')
        
        if isinstance(type, str):
            if not EnhancementElementType.is_valid(type):
                raise Exception(f'The "type" parameter provided {type} is not a valid EnhancementElementType value.')
            
            type = EnhancementElementType(type)

    @classmethod
    def validate_duration(cls, duration: Union[EnhancementElementDuration, int, float, str]):
        """
        This method validates if the provided 'duration' parameter is a valid
        and positive numeric value, if it is a EnhancementElementDuration
        enum or if it is a valid string value of a EnhancementElementDuration
        enum.

        This method will raise an Exception if any of the rules are not fit.
        """
        # 'duration' can be a EnhancementElementDuration enum, a positive
        # numeric value or a string that represents the string value of an
        # EnhancementElementDuration
        if not duration and duration != 0 and duration != 0.0:
            raise Exception('No "duration" provided.')
        
        if not isinstance(duration, EnhancementElementDuration):
            if isinstance(str, duration):
                if not EnhancementElementDuration.is_valid(duration):
                    raise Exception(f'The "duration" parameter provided {str(duration)} is not a valid EnhancementElementDuration value.')
            elif not variable_is_positive_number(duration):
                raise Exception(f'The "duration" parameter provided {str(duration)} is not a positive number.')
            
    @classmethod
    def validate_start(cls, start: Union[EnhancementElementStart, int, float, str]):
        """
        This method validates if the provided 'start' parameter is a valid
        and positive numeric value, if it is a EnhancementElementStart enum
        or if it is a valid string value of a EnhancementElementStart enum.

        This method will raise an Exception if any of the rules are not fit.
        """
        # 'start' can be a EnhancementElementStart enum, a positive
        # numeric value or a string that represents the string value
        # of an EnhancementElementStart
        if not start and start != 0 and start != 0.0:
            raise Exception('No "start" provided.')
        
        if not isinstance(start, EnhancementElementStart):
            if isinstance(start, str):
                if not EnhancementElementStart.is_valid(start):
                    raise Exception(f'The "start" parameter provided {str(start)} is not a valid EnhancementElementStart value.')
                elif not variable_is_positive_number(start):
                    raise Exception(f'The "start" parameter provided {str(start)} is not a positive number.')
                
                start = EnhancementElementStart(start)

        return start
                
    @classmethod
    def validate_mode(cls, mode: Union[EnhancementElementMode, str], valid_modes: Union[EnhancementElementMode, str] = []):
        """
        This method will check if the provided 'mode' is an
        EnhancementElementMode enum or if it is a valid string 
        value of a EnhancementElementMode enum. It will also
        check, if the 'valid_modes' parameter is provided, if
        the provided 'mode' is one of the provided 'valid_modes'.
        If you don't provide the 'valid_modes' this condition 
        will not be checked.

        This method will raise an Exception if any of the rules 
        are not fit.
        """
        if not mode:
            raise Exception('No "mode" provided.')
        
        if not isinstance(mode, (EnhancementElementMode, str)):
            raise Exception(f'The "mode" parameter provided {mode} is not a EnhancementElementMode nor a string.')
        
        if isinstance(mode, str):
            if not EnhancementElementMode.is_valid(mode):
                raise Exception(f'The "mode" provided string {mode} is not a valid EnhancementElementMode enum value.')
            
            mode = EnhancementElementMode(mode)

        if valid_modes and len(valid_modes) > 0:
            if isinstance(valid_modes[0], EnhancementElementMode):
                if not mode in valid_modes:
                    raise Exception(f'The "mode" parameter provided {str(mode)} is not a valid EnhancementElementMode enum.')
            elif isinstance(valid_modes[0], str):
                if not mode.value in valid_modes:
                    raise Exception(f'The "mode" parameter provided {str(mode)} is not a valid value of an EnhancementElement enum.')
            else:
                raise Exception('The provided "valid_modes" are not EnhancementElementMode nor string items.')


from youtube_autonomous.segments.enhancement.edition_manual.enums import EditionManualTermMode, EditionManualTermContext
from youtube_autonomous.segments.enums import EnhancementElementStart, EnhancementElementDuration
from youtube_autonomous.elements.validator.element_validator import StringDuration
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.enhancement.enhancement_element import EnhancementElement
from youtube_autonomous.elements.enhancement import Enhancement
from yta_general_utils.text.transformer import remove_marks_and_accents
from typing import Union
from copy import copy
from re import finditer as re_finditer


class EditionManualTerm:
    """
    A term of the edition manual terms book that is used to look
    for matches in the segments text in order to enhance the video
    or audio content.

    See: https://www.notion.so/Diccionarios-de-mejora-155efcba8f0d44e0890b178effb3be84?pvs=4
    """
    @property
    def term(self):
        # TODO: Explain the term
        return self._term
    
    @term.setter
    def term(self, term: str):
        if not term:
            raise Exception('no "term" provided.')
        
        if not isinstance(term, str):
            raise Exception('The "term" parameter provided is not a string.')
        
        self._term = term

    @property
    def mode(self):
        # TODO: Explain the term
        return self._mode
        
    @mode.setter
    def mode(self, mode: Union[EditionManualTermMode, str]):
        if not mode:
            raise Exception('No "mode" provided.')
        
        if not isinstance(mode, (EditionManualTermMode, str)):
            raise Exception('The "mode" parameter provided is not a valid EditionManualTermMode nor a string.')
        
        if isinstance(mode, str):
            if not EditionManualTermMode.is_valid(mode):
                raise Exception('The "mode" parameter provided is not a valid EditionManualTermMode value.')
            
            mode = EditionManualTermMode(mode)

        self._mode = mode

    @property
    def context(self):
        # TODO: Explain the term
        return self._context
    
    @context.setter
    def context(self, context: Union[EditionManualTermContext, str]):
        if not context:
            raise Exception('No "context" provided.')
        
        if not isinstance(context, (EditionManualTermContext, str)):
            raise Exception('The "context" parameter provided is not a valid EditionManualTermContext nor a string.')
        
        if isinstance(context, str):
            if not EditionManualTermContext.is_valid(context):
                raise Exception('The "context" parameter provided is not a valid EditionManualTermContext value.')
            
            context = EditionManualTermContext(context)

        self._context = context

    @property
    def enhancements(self):
        return self._enhancements
    
    @enhancements.setter
    def enhancements(self, enhancements: list[Enhancement, dict]):
        # TODO: Make some checkings and improvements
        if not enhancements:
            enhancements = []

        # These enhancements should have been validated before when
        # the EditionManual is accepted, so now we consider them as
        # valids

        # Here 'enhancements' are only dicts

        # # We turn dicts to EnhancementElement if necessary
        # obj_enhancements = []
        # for enhancement in enhancements:
        #     if not isinstance(enhancement, EnhancementElement) and not issubclass(enhancement.__class__, EnhancementElement):
        #         obj_enhancements.append(EnhancementElement.get_class_from_type(enhancement['type'])(enhancement['type'], EnhancementElementStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD, EnhancementElementDuration.SHORTCODE_CONTENT, enhancement['keywords'], enhancement.get('url', ''), enhancement.get('filename', ''), enhancement['mode']))
        #     else:
        #         obj_enhancements.append(enhancement)

        self._enhancements = enhancements

    def __init__(self, term: str, mode: str, context: str, enhancements: list[dict]):
        self.term = term
        self.mode = mode
        self.context = context
        self.enhancements = enhancements

    @staticmethod
    def init_from_dict(dict: dict):
        """
        Validates the provided 'dict' that must be a 'key:dict' pair 
        containing the term and its structure. It will raise an 
        Exception if something is wrong or will return a
        EditionManualTerm if everything is ok.
        """
        ElementParameterValidator.validate_edition_manual_term(dict)

        term, dict = dict
        
        return EditionManualTerm(term, dict['mode'], dict['context'], dict['enhancements'])
    
    def search(self, transcription: list[dict]):
        """
        Searches the term in the provided 'transcription' text and, if
        found, returns the corresponding Enhancements with the processed
        duration (if it is possible to process it).

        This method returns the list of enhancement elements that should
        be applied, according to the provided 'transcription', as dict
        elements.
        """
        ElementParameterValidator.validate_transcription_parameter(transcription)

        text = ' '.join(transcription_word['text'] for transcription_word in transcription)
        term = self.term

        # Adapt 'term' and 'text' to search mode
        if self.mode == EditionManualTermMode.IGNORE_CASE_AND_ACCENTS:
            text = remove_marks_and_accents(text).lower()
            term = remove_marks_and_accents(self.term).lower()

        enhancements_found = []
        term_splitted = term.split(' ')
        for found in re_finditer(term, text):
            start_index = found.start()
            subtext = text[:start_index]
            match_first_word_index = len(subtext.split(' ')) - 1
            match_last_word_index = match_first_word_index + len(term_splitted) - 1
            # Here I have the words indexes so, with transcription
            # I can obtain the time start and calculate the duration
            for term_enhancement in self.enhancements:
                enhancement_found = copy(term_enhancement)

                # TODO: By now the only accepted values for EditionManualTerm
                # are these below:
                #   start:
                # EnhancementElementStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD
                # EnhancementElementStart.MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD
                # EnhancementElementStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD
                #   duration:
                # EnhancementElementDuration.SHORTCODE_CONTENT:

                # If only one term, we cannot use this strategy below so we 
                # replace it
                if enhancement_found['start'] == EnhancementElementStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD.value and match_first_word_index == match_last_word_index:
                    enhancement_found['start'] = EnhancementElementStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD.value

                # Here we need to transform string 'start' and 'duration'
                # into their real numeric values
                start = None
                duration = None

                if enhancement_found['start'] == EnhancementElementStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD.value:
                    start = transcription[match_first_word_index]['start']
                    duration = transcription[match_last_word_index]['end'] - start
                elif enhancement_found['start'] == EnhancementElementStart.MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD.value:
                    start = (transcription[match_first_word_index]['start'] + transcription[match_first_word_index]['end']) / 2
                    duration = (transcription[match_last_word_index]['start'] + transcription[match_last_word_index]['end']) / 2 - start
                elif enhancement_found['start'] == EnhancementElementStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD.value:
                    start = transcription[match_first_word_index]['end']
                    duration = transcription[match_last_word_index]['start']

                if start is None or duration is None:
                    raise Exception('Something went wrong when applying a ManualEditionTerm.')

                enhancement_found['start'] = start
                enhancement_found['duration'] = duration
                
                # TODO: Should I turn this into a Enhancement object (?)
                enhancements_found.append(enhancement_found)

        return enhancements_found


"""
"Lionel Messi": {
    "mode": "exact",
    "context": "generic",
    "enhancements": [
        {
            "type": "sticker",
            "keywords": "lionel messi portrait",
            "url": "",
            "filename": "",
            "mode": "overlay"
        }
    ]
}
"""


from youtube_autonomous.segments.enums import EnhancementField, ElementBuildingField, EnhancementOrigin
from youtube_autonomous.shortcodes.objects.shortcode import Shortcode
from youtube_autonomous.elements.element import Element
from yta_general_utils.logger import print_completed
from bson.objectid import ObjectId
from typing import Union


class Enhancement(Element):
    @property
    def segment_index(self):
        return self._segment_index
    
    @segment_index.setter
    def segment_index(self, segment_index: int):
        self._segment_index = segment_index

    @property
    def origin(self):
        return self._origin
    
    @origin.setter
    def origin(self, origin: EnhancementOrigin):
        self._origin = origin

        if self._do_update_database:
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, 'origin', origin)
        # TODO: This should be stored in database

    @property
    def status(self):
        return super().status
    
    @status.setter
    def status(self, status: str):
        super().status = status

        if self._do_update_database:
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, 'status', status)

    @property
    def audio_filename(self):
        return super().audio_filename
    
    @audio_filename.setter
    def audio_filename(self, audio_filename: Union[str, None]):
        audio_filename = audio_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, ElementBuildingField.AUDIO_FILENAME.value, audio_filename)

    @property
    def video_filename(self):
        return super().video_filename
    
    @video_filename.setter
    def video_filename(self, video_filename: Union[str, None]):
        super().video_filename = video_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, ElementBuildingField.VIDEO_FILENAME.value, video_filename)

    @property
    def full_filename(self):
        return super().full_filename
    
    @full_filename.setter
    def full_filename(self, full_filename: Union[str, None]):
        super().full_filename = full_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, ElementBuildingField.FULL_FILENAME.value, full_filename)

    @property
    def audio_narration_filename(self):
        return super().audio_narration_filename
    
    @audio_narration_filename.setter
    def audio_narration_filename(self, audio_narration_filename: Union[str, None]):
        super().audio_narration_filename = audio_narration_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, EnhancementField.AUDIO_NARRATION_FILENAME.value, audio_narration_filename)

    @property
    def transcription(self):
        return super().transcription
    
    @transcription.setter
    def transcription(self, transcription: Union[dict, None]):
        super().transcription = transcription

        if self._do_update_database:
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, ElementBuildingField.TRANSCRIPTION.value, transcription)

    @property
    def shortcodes(self):
        return super().shortcodes
    
    @shortcodes.setter
    def shortcodes(self, shortcodes: Union[list[Shortcode], None]):
        super().shortcodes = shortcodes

        if self._do_update_database:
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, ElementBuildingField.SHORTCODES.value, shortcodes)

    @property
    def start(self):
        return self.data.start
    
    @start.setter
    def start(self, start):
        self.data.start = start

    @property
    def calculated_duration(self):
        return super().calculated_duration
    
    @calculated_duration.setter
    def calculated_duration(self, calculated_duration: Union[float, int, None]):
        super().calculated_duration = calculated_duration

        if self._do_update_database:
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, ElementBuildingField.CALCULATED_DURATION.value, calculated_duration)

    def __init__(self, project_id: Union[str, ObjectId], segment_index: int, index: int, data: dict):
        super().__init__(index, data)

        self._do_update_database = False

        self.project_id = str(project_id)
        self.segment_index = segment_index
        
        # By now, we are not accepting shortcodes nor other
        # enhancements in enhancement elements

        self._do_update_database = True

    # TODO: Remove this method below
    def prepare(self):
        # Check duration
        if self.data.duration and not self.calculated_duration:
            self.calculated_duration = self.data.duration

        # Generate narration audio
        if self.rules_checker.should_build_narration_rule(self.data) and not self.audio_clip:
            self.handle_narration()

        # Generate video
        if not self.video_clip:
            self.video_clip = self.builder.build_segment(self)
            # We write this video part to be able to recover it in the future
            segment_part_filename = self.create_segment_file('video.mp4')
            self.video_clip.write_videofile(segment_part_filename)
            self.video_filename = segment_part_filename
            print_completed('Created basic video part and stored locally and in database.')

    # TODO: This below has to be in the common Object from wich Segment
    # and Enhancement will inherit
    def build_step_1_create_narration(self):
        # 1. Generate narration if needed
        self.create_narration()

    def build_step_2_build_base_content(self, segment):
        # 2. Build base video
        if not self.video_clip:
            self.video_clip = self.builder.build_from_segment(self)
            filename = self.create_segment_file('video.mov')
            self.video_clip.write_videofile(filename)
            self.video_filename = filename

        # TODO: In the future we could add enhancements to this
        if self.audio_clip:
            self.video_clip = self.video_clip.set_audio(self.audio_clip)
        self.full_clip = self.video_clip
        filename = self.create_segment_file('video.mov')
        self.full_clip.write_videofile(filename)
        self.full_filename = filename
    
    def build(self, segment):
        """
        Builds this enhancement to be applied in the provided 'segment'.
        """
        self.build_step_1_create_narration()
        self.build_step_2_build_base_content(segment)

        return self.full_clip


from youtube_autonomous.elements.rules.element_rules import ElementRules
from youtube_autonomous.segments.enums import EnhancementElementMode


class EffectElementRules(ElementRules):
    def __init__(self):
        self.can_have_narration = False
        self.need_narration = False
        self.can_have_specific_duration = False
        self.need_specific_duration = False
        self.can_have_text = False
        self.need_text = False
        self.can_have_filename = False
        self.can_have_url = False
        self.need_filename_or_url = False
        self.can_have_keywords = True
        self.need_keywords = True
        self.can_have_more_parameters = True
        # TODO: The effect should have some way to detect the
        # mandatory parameters so we can dynamically raise an
        # Exception if any of those mandatory parameters are
        # not provided

        self.can_be_segment = False
        self.can_be_enhancement_element = True
        self.valid_enhancement_modes = [EnhancementElementMode.REPLACE]
        self.default_enhancement_mode = EnhancementElementMode.REPLACE



from youtube_autonomous.elements.validator.error_message import ErrorMessage
from youtube_autonomous.segments.enums import SegmentType, EnhancementType, SegmentStatus, ProjectStatus
from youtube_autonomous.segments.enhancement.edition_manual.enums import EditionManualTermContext, EditionManualTermMode
#from youtube_autonomous.elements.rules.effect_element_rules import ElementRules
#from youtube_autonomous.elements.rules.rules_checker import RulesChecker
from youtube_autonomous.elements.validator import RULES_SUBCLASSES, BUILDER_SUBCLASSES
from yta_general_utils.programming.parameter_validator import ParameterValidator
from yta_general_utils.file.filename import FileType
from typing import Union


class ElementParameterValidator(ParameterValidator):
    """
    A validator that will raise an Exception if any of the used
    checkings is invalid, or will return the value if ok.
    """
    @classmethod
    def validate_keywords(cls, keywords: str):
        cls.validate_string_mandatory_parameter('keywords', keywords)

        return keywords

    @classmethod
    def validate_text(cls, text: str):
        cls.validate_string_mandatory_parameter('text', text)

        return text
    
    @classmethod
    def validate_premade_name(cls, premade_name: str):
        cls.validate_string_mandatory_parameter('premade_name', premade_name)
        # TODO: Maybe validate that is a key of a valid premade (?)

        return premade_name
    
    @classmethod
    def validate_text_class_name(cls, text_class_name: str):
        cls.validate_string_mandatory_parameter('text_class_name', text_class_name)
        # TODO: Maybe validate that is a key of a valid text_class (?)

        return text_class_name
    
    @classmethod
    def validate_duration(cls, duration: Union[int, float]):
        # TODO: Maybe validate that is a key of a valid text class (?)
        cls.validate_numeric_positive_mandatory_parameter('duration', duration)

        return duration
    
    @classmethod
    def validate_url(cls, url: str, is_mandatory: bool = True):
        """
        A mandatory 'url' must be not None, a valid string and also
        a valid url and accessible.

        A non mandatory 'url' can be None, 
        """
        if not is_mandatory and not url:
            return True
        
        cls.validate_mandatory_parameter('url', url)
        cls.validate_string_parameter('url', url)
        cls.validate_url_is_ok('url', url)
        # TODO: Validate 'url' file is some FileType

        return url

    @classmethod
    def validate_filename(cls, filename: str, file_type: FileType = FileType.IMAGE, is_mandatory: bool = True):
        """
        Validates the 'filename' parameter. If the 'filename' parameter has
        some value, all conditions will be checked. If there is not 'filename'
        and 'is_mandatory' is False, no Exceptions will be raised because it
        is valid.
        """
        if not is_mandatory and not filename:
            return True
        
        cls.validate_mandatory_parameter('filename', filename)
        cls.validate_string_parameter('filename', filename)
        cls.validate_file_exists('filename', filename)
        if file_type:
            cls.validate_filename_is_type('filename', filename, file_type)
            # TODO: Validate type trying to instantiate it

        return filename

    @classmethod
    def validate_rules(cls, rules: 'ElementRules'):
        print(type(rules).__name__)
        if type(rules).__name__ not in RULES_SUBCLASSES:
            raise Exception(ErrorMessage.parameter_is_not_rules('rules'))
        
        return rules
        
    @classmethod
    def validate_builder(cls, builder: 'ElementBuilder'):
        if type(builder).__name__ not in BUILDER_SUBCLASSES:
            raise Exception(ErrorMessage.parameter_is_not_builder('builder'))
        
        return builder
        
    @classmethod
    def validate_segment_type(cls, type: Union[SegmentType, str]):
        """
        This method validates that the provided 'type' is not None
        and is a valid type and accepted for Segment building.

        This method raises an Exception if something is wrong
        and returns the type as a SegmentType enum if everything
        was ok.
        """
        return cls.validate_enum('type', type, SegmentType)

    @classmethod
    def validate_enhancement_type(cls, type: Union[EnhancementType, str]):
        """
        Validates that the provided 'type' is not None
        and is a valid type and accepted for Enhancement building.

        This method raises an Exception if something is wrong
        and returns the type as a EnhancementType enum if 
        everything was ok.
        """
        return cls.validate_enum('type', type, EnhancementType)
    
    @classmethod
    def validate_segment_or_enhancement_type(cls, type: Union[SegmentType, EnhancementType, str]):
        """
        Validates if the provided 'type' is a valid SegmentType or
        EnhancementType enum object, or a valid SegmentType or 
        EnhancementType enum value, raising an Exception if not.

        This method will return the type as the corresponding 
        enum if everything is ok.
        """
        enum_type = None
        try:
            enum_type = cls.validate_segment_type(type)
        except Exception:
            pass

        if not enum_type:
            try:
                enum_type = cls.validate_enhancement_type(type)
            except Exception:
                pass

        if not enum_type:
            raise Exception(f'The "type" parameter provided {type} is not a valid SegmentType or EnhancementType.')
        
        return enum_type

    @classmethod
    def validate_segment_status(cls, status: Union[SegmentStatus, str]):
        """
        Validates if the provided 'status' is not None and is
        a valid SegmentStatus Enum or string value.

        This method raises an Exception if something is wrong
        and returns the status as a SegmentStatus enum if 
        everything was ok.
        """
        return cls.validate_enum('status', status, SegmentStatus)
    
    @classmethod
    def validate_project_status(cls, status: Union[ProjectStatus, str]):
        """
        Validates if the provided 'status' is not None and is
        a valid SegmentStatus Enum or string value.

        This method raises an Exception if something is wrong
        and returns the status as a ProjectStatus enum if 
        everything was ok.
        """
        return cls.validate_enum('status', status, ProjectStatus)
    
    @classmethod
    def validate_edition_manual_term_mode(cls, mode: Union[EditionManualTermMode, str]):
        """
        Validates if the provided 'mode' is not None and is
        a valid EditionManualTermMode Enum or string value.

        This method raises an Exception if something is wrong
        and returns the mode as an EditionManualTermMode enum 
        if everything was ok.
        """
        return cls.validate_enum('mode', mode, EditionManualTermMode)
        
    @classmethod
    def validate_edition_manual_term_context(cls, context: Union[EditionManualTermContext, str]):
        """
        Validates if the provided 'context' is not None and
        is a valid EditionManualTermContext Enum or string 
        value.

        This method raises an Exception if something is wrong
        and returns the mode as an EditionManualTermContext 
        enum if everything was ok.
        """
        return cls.validate_enum('context', context, EditionManualTermContext)
    
    @classmethod
    def validate_edition_manual_term(cls, edition_manual_term: dict):
        """
        This method verifies if the provided 'dict' has a valid
        structure for an EditionManualTerm, which has to be a key
        and a dict value containing all the required fields.

        This method will raise an Exception if something is wrong.
        """
        #_, dict = edition_manual_term
        term, term_content = next(iter(edition_manual_term.items()))

        # TODO: Apply Enums
        ElementParameterValidator.validate_edition_manual_term_mode(term_content.get('mode', None))
        ElementParameterValidator.validate_edition_manual_term_context(term_content.get('context', None))

        enhancements = term_content.get('enhancements', None)
        if enhancements is None:
            raise Exception('No "enhancements" field found in the provided "edition_manual_term".')
        
        if len(enhancements) == 0:
            raise Exception('The "enhancements" field is empty.')

        # TODO: Cyclic import error
        # for enhancement in enhancements:
        #     ElementParameterValidator.validate_segment_or_enhancement(enhancement)

    @classmethod
    def validate_transcription_parameter(cls, transcription: list[dict]):
        # for transcription_word in transcription:
        #     if not 'word' in transcription_word or not 'start' in transcription_word or not 'end' in transcription_word:
        #         raise Exception('At least one term of the provided "transciption" parameter has no "word", "start" or "end" field.')
        if any(not 'text' in transcription_word or not 'start' in transcription_word or not 'end' in transcription_word for transcription_word in transcription):
            raise Exception('At least one term of the provided "transcription" parameter has no "text", "start" or "end" field.')


from youtube_autonomous.segments.enhancement.edition_manual.edition_manual_term import EditionManualTerm
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from yta_general_utils.file.reader import read_json_from_file
from typing import Union


class EditionManual:
    @property
    def terms(self):
        return self._terms
    
    @terms.setter
    def terms(self, terms: list[EditionManualTerm]):
        if not terms:
            terms = []

        if any(not isinstance(term, EditionManualTerm) for term in terms):
            raise Exception('At least one of the provided "terms" is not an EditionManualTerm.')

        self._terms = terms

    @staticmethod
    def init_from_file(filename: str):
        """
        Initializes an Edition Manual from the given file
        'filename' that must have a valid structure.
        """
        ElementParameterValidator.validate_filename(filename, None)

        terms_dictionary = read_json_from_file(filename)

        terms = terms_dictionary.get('terms', None)

        if terms is None:
            raise Exception('No "terms" field found in the provided json filename {filename}.')
        
        # Terms will be validated in the '__init__' method

        return EditionManual(terms)

    def __init__(self, terms: Union[list[dict], list[EditionManualTerm]]):
        # TODO: We need an ID or something maybe
        # TODO: Look for a better way to iterate
        term_objects = []
        for term in terms:
            if not isinstance(term, EditionManualTerm):
                obj = {
                    term: terms[term]
                }
                ElementParameterValidator.validate_edition_manual_term(obj)
                term_objects.append(EditionManualTerm(term, terms[term]['mode'], terms[term]['context'], terms[term]['enhancements']))
            else:
                term_objects.append(term)

        self.terms = term_objects

    def apply(self, transcription: list[dict]):
        """
        Applies this edition manual to the provided 'segment' parameter.
        This method will search all the terms in the edition manual and
        will return all the enhancements found according to those terms
        based on the segment transcription.
        """
        # We only accept segments by now, enhancements in a future
        
        # Search for the terms according to the search mode
        # of each term
        enhancements_found = []
        for term in self.terms:
            enhancements_found += term.search(transcription)

        # Enhancements are, by now, dicts
        # TODO: Maybe we should append enhancements as objects to
        # the segment
        return enhancements_found


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from yta_multimedia.greenscreen.custom.image_greenscreen import ImageGreenscreen
from yta_multimedia.greenscreen.custom.video_greenscreen import VideoGreenscreen
from moviepy.editor import VideoFileClip, CompositeVideoClip, ColorClip, ImageClip, AudioFileClip, AudioClip, CompositeAudioClip
from typing import Union


class GreenscreenElementBuilder(ElementBuilder):
    """
    This builders allows you to generate 'GREENSCREEN' content.
    """
    @classmethod
    def build_from_enhancement(cls, enhancement: dict):
        filename = enhancement.filename
        url = enhancement.url
        # TODO: I need the 'segment' video to apply the greenscreen
        video = None

        if filename:
            # TODO: Check if 'image' or 'video'
            is_video = True
            # TODO: Build with filename
            if is_video:
                return cls.build_with_video_greenscreen(filename, video)
            return cls.build_with_image_greenscreen(filename, video)
        # TODO: Check if 'image' or 'video'
        is_video = True
        # TODO: Build with url
        if is_video:
            return cls.build_with_video_greenscreen(url, video)
        return cls.build_with_image_greenscreen(url, video)

    # TODO: Remove this method below when the other one is dynamic
    @classmethod
    def build_with_video_greenscreen(cls, filename_or_google_drive_url: str, video_or_audio: Union[VideoFileClip, CompositeVideoClip, ColorClip, ImageClip, AudioFileClip, AudioClip, CompositeAudioClip]):
        ElementParameterValidator.validate_string_mandatory_parameter('filename_or_google_drive_url', filename_or_google_drive_url)
        # TODO: Validate it is filename or url
        # TODO: Check if it is video or image to dynamically use 
        # the ImageGreenscreen or the VideoGreenscreen
        return VideoGreenscreen(filename_or_google_drive_url).from_video_to_video(video_or_audio)

    @classmethod
    def build_with_image_greenscreen(cls, filename_or_google_drive_url: str, video_or_audio: Union[VideoFileClip, CompositeVideoClip, ColorClip, ImageClip, AudioFileClip, AudioClip, CompositeAudioClip]):
        ElementParameterValidator.validate_string_mandatory_parameter('filename_or_google_drive_url', filename_or_google_drive_url)
        # TODO: Validate it is filename or url
        # TODO: Check if it is video or image to dynamically use 
        # the ImageGreenscreen or the VideoGreenscreen
        return ImageGreenscreen(filename_or_google_drive_url).from_video_to_video(video_or_audio)

    @classmethod
    def build(cls, video_or_audio: Union[VideoFileClip, CompositeVideoClip, ColorClip, ImageClip, AudioFileClip, AudioClip, CompositeAudioClip]):
        """
        Basic example to test that the building process and
        the class are working correctly.

        TODO: Remove this in the future when 'custom' is 
        working perfectly.
        """
        return ImageGreenscreen('https://drive.google.com/file/d/1WQVnXY1mrw-quVXOqTBJm8x9scEO_JNz/view?usp=sharing').from_video_to_video(video_or_audio)
        # return VideoGreenscreen('https://drive.google.com/file/d/1hL677Q87jn-_y5vqSwev3lir_dCYRQwS/view?usp=sharing').apply()


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.enums import SegmentField
from yta_multimedia.video.generation.manim.classes.text.text_word_by_word_manim_animation import TextWordByWordManimAnimation
from moviepy.editor import VideoFileClip
from typing import Union


class TextElementBuilder(ElementBuilder):
    @classmethod
    def build_custom_from_text_class_name(cls, text_class_name: str, **parameters):
        ElementParameterValidator.validate_text_class_name(text_class_name)

        text_class = None
        if text_class == 'text_word_by_word':
            text_class = TextWordByWordManimAnimation
        else:
            raise Exception(f'The provided "text_class" parameter {text_class} is not a valid text class name.')
        
        return cls.build_custom(text_class, **parameters)


    @classmethod
    def build_custom(cls, text_animation_class, **parameters):
        """
        This method instantiates the 'text_animation_class' Manim
        text animation class and uses the provided 'parameters' to
        build the text animation. The provided 'parameters' must 
        fit the ones requested by the provided class 'generate'
        method.
        """
        raise Exception('Not implemented yet.')
        # TODO: Apply TextManimAnimation to 'text_animatino_class'
        # TODO: Check that 'text_animation_class' is subclass of 
        # abstract TextManimAnimation (that I need to create)
        # TODO: Check that I have the required parameters. Is
        # this possible (?)
        filename = text_animation_class().generate(**parameters)

        return VideoFileClip(filename)

    @classmethod
    def build_from_enhancement(cls, enhancement: dict):
        # TODO: This parameter must be an enum and, is this the definitive
        # one (?)
        # TODO: If narration, the transcription should be here, if no 
        # narration, the text
        text = enhancement.get('text', None)
        # TODO: I should always have 'calculated_duration' when duration
        # has been processed
        duration = enhancement.get('calculated_duration', None)

        return cls.build(text, duration)

    @classmethod
    def build_from_segment(cls, segment: dict):
        text = segment.get(SegmentField.TEXT.value, None)
        duration = segment.get(SegmentField.DURATION.value, None)

        return cls.build(text, duration)

    @classmethod
    def build(cls, text: str, duration: Union[int, float]):
        """
        Basic example to test that the building process and
        the class are working correctly.

        TODO: Remove this in the future when 'custom' is 
        working perfectly.
        """
        ElementParameterValidator.validate_text(text)
        ElementParameterValidator.validate_duration(duration)

        filename = TextWordByWordManimAnimation().generate(text, duration)

        video = VideoFileClip(filename)

        return video


from youtube_autonomous.segments.enhancement.enhancement_element import EnhancementElement, EnhancementElementMode
from yta_multimedia.video.utils import parse_parameter_as_moviepy_clip
from yta_multimedia.video.generation import generate_video_from_image
from yta_general_utils.downloader.image import download_image
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip
from typing import Union


class ImageEnhancementElement(EnhancementElement):
    @classmethod
    def get_valid_modes(cls):
        """
        The EnhancementElementMode enums accepted for this type of
        EnhancementElement.

        See this link: https://www.notion.so/Segment-types-configuration-107f5a32d46280beb793d84a2af7f75e?pvs=4
        """
        return [EnhancementElementMode.INLINE, EnhancementElementMode.OVERLAY]
    
    @classmethod
    def get_default_mode(cls):
        """
        The EnhancementElementMode enum defined as value by default.
        """
        return EnhancementElementMode.OVERLAY

    @property
    def valid_modes(self):
        """
        The EnhancementElementMode enums accepted for this type of
        EnhancementElement.

        See this link: https://www.notion.so/Segment-types-configuration-107f5a32d46280beb793d84a2af7f75e?pvs=4
        """
        return ImageEnhancementElement.get_valid_modes()
    
    def _build(self, video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, str]):
        filename = ''
        if self.filename:
            filename = self.filename
        elif self.url:
            filename = download_image(self.url)
        elif self.keywords:
            # TODO: Search image and download it from somewhere
            filename = download_image('https://png.pngtree.com/thumb_back/fh260/background/20230611/pngtree-wolf-animals-images-wallpaper-for-pc-384x480-image_2916211.jpg')

        enhancement_clip = generate_video_from_image(filename, self.duration)

        super().recalculate_duration(enhancement_clip, video)

        return enhancement_clip

    def apply(self, video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, str]):
        """
        This method will apply this Enhancement Element in the provided
        'video' and will return it modified including the generated
        enhancement clip.
        """
        video = parse_parameter_as_moviepy_clip(video)

        enhancement_clip = self._build(video)

        video = super().apply(enhancement_clip, video)

        return video